from dataclasses import dataclass
import shutil
from typing import List, Literal

from bygg.action import Action
from bygg.common_types import CommandStatus, JobStatus, Severity
from bygg.output import (
    Symbols,
    output_error,
    output_info,
    output_warning,
    output_with_status_line,
)


def on_runner_status(message: str):
    output_info(message)


running_jobs: set[str] = set()


def format_queued_jobs_line() -> str:
    terminal_cols, _ = shutil.get_terminal_size()
    output = f"{' '.join(running_jobs)}"
    if len(output) > terminal_cols:
        output = output[: terminal_cols - 3] + "..."
    return output


max_name_length = 0


def print_job_ended(
    name: str, job_status: JobStatus, action: Action, status: CommandStatus | None
):
    global max_name_length
    max_name_length = max(len(name), max_name_length)
    failed_or_stopped = job_status in ("failed", "stopped")
    symbol = Symbols.RED_X if failed_or_stopped else Symbols.GREEN_CHECKMARK
    status_code_message = f"[{status.rc}] " if status else "?"
    status_message = status.message if status and status.message else ""
    output_with_status_line(
        format_queued_jobs_line(),
        f"{symbol} {name:<{max_name_length}} : {status_code_message if status and status.rc else ''}{status_message}",
    )


def on_job_status(
    name: str, job_status: JobStatus, action: Action, status: CommandStatus | None
):
    match job_status:
        case "skipped":
            pass
        case "running":
            running_jobs.add(name)
        case s if s in ("failed", "finished", "stopped"):
            running_jobs.discard(name)
            print_job_ended(name, job_status, action, status)
        case _:
            raise ValueError(f"Unhandled job status {job_status}")


CheckRule = Literal["check_inputs_outputs", "output_file_missing"]
"""
The different rules that can be checked.

    check_inputs_outputs: Check that earlier actions don't need outputs from subsequent
    actions.

    output_file_missing: Check that actions create the files that they declare as
    outputs.
"""


@dataclass
class CheckStatus:
    rule_name: CheckRule
    action: Action
    status_text: str
    severity: Severity


failed_checks: List[CheckStatus] = []


def on_check_failed(
    rule_name: CheckRule, action: Action, status_text: str, severity: Severity
):
    failed_checks.append(CheckStatus(rule_name, action, status_text, severity))


def output_check_results():
    status = True
    output_error("The following checks reported issues:")
    for c in failed_checks:
        compound_status = f"{c.rule_name} :: {c.action.name} :: {c.status_text}"
        match c.severity:
            case "error":
                status = False
                output_error(compound_status)
            case "warning":
                status = False
                output_warning(compound_status)
            case "info":
                continue
            case _:
                raise ValueError(f"Unhandled severity {c.severity}")

    return status



from itertools import chain
from typing import List

from bygg.output import TerminalStyle as TS
from bygg.scheduler import Scheduler


class TreeStyle:
    """Base class for tree styles."""

    BAR = "─"
    PIPE = "│"
    T_JOINT = "├"
    END_CORNER = "└"
    CONNECTOR = T_JOINT + BAR * 2 + " "
    HANGER = END_CORNER + BAR * 2 + " "


class TreeStyleUnicode(TreeStyle):
    """Same style as base class, but with parametrized indent."""

    def __init__(self, indent=4):
        self.CONNECTOR = (
            f"{self.T_JOINT + self.BAR * (indent - 1 - len(self.T_JOINT)):<{indent}}"
        )
        self.HANGER = f"{self.END_CORNER + self.BAR * (indent - 1 - len(self.END_CORNER)):<{indent}}"


class TreeStyleAscii(TreeStyle):
    """ASCII characters instead of unicode."""

    BAR = "-"
    PIPE = "|"
    T_JOINT = "+"
    END_CORNER = "\\"

    def __init__(self, indent=4):
        self.CONNECTOR = (
            f"{self.T_JOINT + self.BAR * (indent - 1 - len(self.T_JOINT)):<{indent}}"
        )
        self.HANGER = f"{self.END_CORNER + self.BAR * (indent - 1 - len(self.END_CORNER)):<{indent}}"


def display_tree(scheduler: Scheduler, entry_points: List[str]):
    """
    Display the dependency tree for the given entry points.

    Example output:

    all_checks
    ├── check_inputs_outputs
    │   └── circular_C
    │       └── circular_B
    │           └── circular_A
    └── output_file_missing
        └── no_outputs_A
    """

    indent = 4
    style = TreeStyleUnicode(indent)

    for entry_point in entry_points:
        build_actions = scheduler.build_actions

        def format_children(name: str, last_sibling: bool, depth: int) -> List[str]:
            action = build_actions[name]
            display_name = f"{TS.BOLD}{name}{TS.RESET}" if depth == 0 else name

            # Format children and flatten:
            children = chain.from_iterable(
                (
                    format_children(dep, i == len(action.dependencies) - 1, depth + 1)
                    for i, dep in enumerate(sorted(action.dependencies))
                )
            )

            # Indent with the correct prefixes:
            prefix = style.HANGER if last_sibling else style.CONNECTOR
            subtree = [f"{prefix if depth > 0 else ''}{display_name}"]

            child_prefix = f"{style.PIPE if not last_sibling else ' ':<{indent}}"
            subtree.extend(
                [f"{child_prefix if depth > 0 else ''}{item}" for item in children]
            )
            return subtree

        print()
        print("\n".join(format_children(entry_point, True, 0)))
    return True



from bygg.action import Action
from bygg.util import create_shell_command, filenames_from_pattern
import pytest

test_cases_filenames_from_pattern = [
    (
        (
            ["/foo/bar/baz.txt.j2"],
            "%.j2",
            "out_%",
            [("/foo/bar/baz.txt.j2", "/foo/bar/out_baz.txt")],
        )
    ),
    (
        (
            ["/foo/bar/baz.txt.j2"],
            "%.txt.j2",
            "out_%.txt",
            [("/foo/bar/baz.txt.j2", "/foo/bar/out_baz.txt")],
        )
    ),
    (
        (
            ["foo.c", "bar.c"],
            "%.c",
            "%.o",
            [("foo.c", "foo.o"), ("bar.c", "bar.o")],
        )
    ),
    (
        (
            ["foo.txt", "bar.txt"],
            "%",
            "out_%",
            [("foo.txt", "out_foo.txt"), ("bar.txt", "out_bar.txt")],
        )
    ),
]


def construct_test_case_name(x: tuple):
    a, b, c, d = x
    return f"{b} | {c}"


@pytest.mark.parametrize(
    "test_case", test_cases_filenames_from_pattern, ids=construct_test_case_name
)
def test_filenames_from_pattern(test_case):
    input_files, in_pattern, out_pattern, output_files = test_case
    assert filenames_from_pattern(input_files, in_pattern, out_pattern) == output_files


def test_shell_command():
    shell_command = create_shell_command("echo 'shell action output'")

    status = shell_command(Action("test context", inputs=set(), outputs=set()))
    assert status.rc == 0
    assert status.output == "shell action output\n"



from bygg.tree import display_tree


def test_tree_single_action(scheduler_single_action, capsys, snapshot):
    scheduler, _ = scheduler_single_action
    display_tree(scheduler, ["action1"])
    stdout, _ = capsys.readouterr()
    assert stdout == snapshot


def test_tree_four_nonbranching_actions(
    scheduler_four_nonbranching_actions, capsys, snapshot
):
    scheduler, _ = scheduler_four_nonbranching_actions
    display_tree(scheduler, ["action1"])
    stdout, _ = capsys.readouterr()
    assert stdout == snapshot


def test_tree_branching_actions(scheduler_branching_actions, capsys, snapshot):
    scheduler, _ = scheduler_branching_actions
    display_tree(scheduler, ["action1"])
    stdout, _ = capsys.readouterr()
    assert stdout == snapshot



from abc import ABCMeta, abstractmethod
from collections import deque
from typing import Any, Dict, Iterable, List

from bygg.action import Action


class Dag(metaclass=ABCMeta):
    def __len__(self) -> int:
        return 0

    def clear(self):
        ...

    def remove_node(self, node: str):
        ...

    def build_action_graph(self, build_actions: Dict[str, Action], action: Action):
        ...

    @abstractmethod
    def get_ready_jobs(
        self,
        finished_jobs: Dict[str, Any],
        running_jobs: Dict[str, Any],
    ) -> List[str]:
        ...

    @abstractmethod
    def get_all_jobs(self) -> Iterable[str]:
        ...


class GtDag(Dag):
    def __init__(self):
        import graph  # type: ignore

        self.graph = graph.Graph()

    def __len__(self):
        nodes = self.graph.nodes()
        if nodes is None:
            return 0
        return len(nodes)

    def clear(self):
        nodes = self.graph.nodes()
        if nodes is None:
            return
        for n in nodes:
            self.graph.del_node(n)

    def remove_node(self, node: str):
        self.graph.del_node(node)

    def build_action_graph(self, build_actions: Dict[str, Action], action: Action):
        """Build the action graph."""
        queue = deque([action])
        while len(queue) > 0:
            a = queue.popleft()
            self.graph.add_node(a.name)
            for dependency in a.dependencies:
                dependency_action = build_actions.get(dependency)
                if not dependency_action:
                    raise ValueError(f"Action '{dependency}' not found")
                queue.append(dependency_action)
                self.graph.add_edge(a.name, dependency)

    def get_ready_jobs(
        self,
        finished_jobs: Dict[str, Any],
        running_jobs: Dict[str, Any],
    ) -> List[str]:
        nodes = self.graph.nodes()
        if nodes is None:
            return []
        if len(nodes) == 1 and (
            nodes[0] not in finished_jobs and nodes[0] not in running_jobs
        ):
            return nodes

        new_nodes = self.graph.nodes(out_degree=0)
        if new_nodes is None:
            return []

        new_jobs = [
            n for n in new_nodes if n not in finished_jobs and n not in running_jobs
        ]

        # print(f"new jobs: {len(new_jobs)}")
        return new_jobs

    def get_all_jobs(self) -> Iterable[str]:
        nodes = self.graph.nodes()
        if nodes is None:
            return []
        return nodes


def create_dag() -> Dag:
    return GtDag()



from pathlib import Path

from bygg.completions import ByggCompletionFinder
from bygg.main import create_argument_parser
from bygg.system_helpers import change_dir
import pytest


def dummy_exit_method(status):
    pass


def open_raise(*args, **kwargs):
    raise NotImplementedError


@pytest.fixture
def completion_tester(monkeypatch, tmp_path):
    def tester(testcase):
        prefix = "bygg"
        prefixed_testcase = f"{prefix} {testcase}"
        outfile = tmp_path / "completion.out"

        # Simulate being run from the completion script:
        monkeypatch.setenv("_ARGCOMPLETE", "1")
        monkeypatch.setenv("COMP_LINE", prefixed_testcase)
        monkeypatch.setenv("COMP_POINT", str(len(prefixed_testcase)))
        monkeypatch.setenv("_ARGCOMPLETE_STDOUT_FILENAME", str(outfile))
        # Patch os.fdopen to raise NotImplementedError; this causes argcomplete to use
        # stderr for debug output instead of opening file descriptor 9, which causes pytest
        # to get its knickers in a twist.
        monkeypatch.setattr("os.fdopen", open_raise)

        parser = create_argument_parser()
        completer = ByggCompletionFinder()

        with open(outfile, "w", encoding="utf-8"):
            completer(parser, exit_method=dummy_exit_method)

        with open(outfile, "r", encoding="utf-8") as f:
            return f.read()

    return tester


args = [
    # Complete simple arguments:
    ("--ver", "--version "),
    ("--cl", "--clean "),
    ("--h", "--help "),
    # No completions after these arguments:
    ("-h ", ""),
    ("--help ", ""),
    ("-v ", ""),
    ("--version ", ""),
    ("--dump-schema ", ""),
    ("--completions ", ""),
]


@pytest.mark.parametrize("arg", args, ids=lambda x: x[0])
def test_completions(completion_tester, arg):
    testcase, correct_result = arg
    testresult = completion_tester(testcase)
    assert correct_result == testresult


# Directories:
directories_completions = [
    "-C ",
    "--directory ",
]


@pytest.mark.parametrize("arg", directories_completions, ids=lambda x: x)
def test_directory_completions(completion_tester, arg, snapshot):
    testcase = arg
    testresult = completion_tester(testcase)
    assert sorted(testresult.split("\x0b")) == snapshot


actions_completions = [p for p in Path("examples").iterdir() if p.is_dir()]


@pytest.mark.parametrize("arg", actions_completions, ids=lambda x: str(x.name))
def test_actions_completions(completion_tester, arg, snapshot, clean_bygg_tree):
    with change_dir(clean_bygg_tree):
        dir = str(arg)
        testcase = f"-C {dir} "
        testresult = completion_tester(testcase)
        assert sorted(testresult.split("\x0b")) == snapshot

        testcase = f"--directory {dir} "
        testresult = completion_tester(testcase)
        assert sorted(testresult.split("\x0b")) == snapshot

        with change_dir(dir):
            testresult = completion_tester("")
            assert sorted(testresult.split("\x0b")) == snapshot



import os
import sys
from typing import Dict, List, Optional

import msgspec

from bygg.output import Symbols, output_plain
from bygg.output import TerminalStyle as TS

PYTHON_INPUTFILE = "Byggfile.py"
YAML_INPUTFILE = "Byggfile.yml"


class Settings(msgspec.Struct, forbid_unknown_fields=True):
    default_action: Optional[str] = None


class ActionItem(msgspec.Struct, forbid_unknown_fields=True):

    """
    This is a representation of the Action class used for deserialising from YAML.

    Parameters
    ----------
    name : str
        The name of the action.
    description : str, optional
        A description of the action. Used in e.g. action listings. Default is None.
    message : str, optional
        A message to print when the action is executed. Default is None.
    inputs : list of str, optional
        A list of files that are used as input to the action. Default is None.
    outputs : list of str, optional
        A list of files that are generated by the action. Default is None.
    dependencies : list of str, optional
        A list of actions that must be executed before this action. Default is None.
    is_entrypoint : bool, optional
        Whether this action is an entrypoint. Entrypoints are actions that can be
        executed directly from the command line. If not set, this is treated as true by
        default in Byggfile.yml and false when used from Python. Default is None.
    environment : str, optional
        The environment for the action. Default is "default".
    shell : str, optional
        The shell command to execute when the action is run. Default is None.
    """

    name: str
    description: Optional[str] = None
    message: Optional[str] = None
    inputs: Optional[List[str]] = None
    outputs: Optional[List[str]] = None
    dependencies: Optional[List[str]] = None
    is_entrypoint: Optional[bool] = None
    environment: Optional[str] = "default"
    shell: Optional[str] = None


class Environment(msgspec.Struct, forbid_unknown_fields=True):
    """
    A class used to represent a virtual environment that actions can be run in.

    Attributes
    ----------
    byggfile : str
        The Python Byggfile that uses this environment.
    inputs : List[str]
        A list of files that are used as input to the environment. Typically pip
        requirements files, but can be any files.
    venv_directory : str
        The directory where the virtual environment is located. Will be recreated by
        Bygg if any of the inputs are modified.
    shell : str
        The shell command for creating the environment.
    description : str, optional
        A description of the environment. Used in e.g. help messages, by default None
    message : str, optional
        A message related to the environment, by default None
    """

    byggfile: str
    inputs: List[str]
    venv_directory: str
    shell: str
    description: Optional[str] = None
    message: Optional[str] = None


class ByggFile(msgspec.Struct, forbid_unknown_fields=True):
    actions: List[ActionItem]
    settings: Settings = msgspec.field(default_factory=Settings)
    environments: Dict[str, Environment] = msgspec.field(default_factory=dict)


def read_config_file() -> ByggFile:
    if not os.path.isfile(YAML_INPUTFILE):
        return ByggFile(actions=[], settings=Settings(), environments={})

    try:
        with open(YAML_INPUTFILE, "r") as f:
            config_file = msgspec.yaml.decode(f.read(), type=ByggFile)
        for action in config_file.actions:
            if action.is_entrypoint is None:
                action.is_entrypoint = True
        return config_file
    except Exception as e:
        output_plain(
            Symbols.RED_X
            + TS.Fg.RED
            + " Error while reading configuration file "
            + TS.Fg.RESET
            + f"{TS.BOLD}{YAML_INPUTFILE}{TS.RESET}. {e}"
        )
        sys.exit(1)


def dump_schema():
    import json

    schema = msgspec.json.schema(ByggFile)
    print(json.dumps(schema, indent=2))



from dataclasses import dataclass
from pathlib import Path
import subprocess

import pytest


@dataclass
class ExampleParameters:
    name: str
    list_rc: int = 0
    tree_rc: int = 0


examples = [
    ExampleParameters("checks"),
    ExampleParameters("environments"),
    ExampleParameters("failing_jobs"),
    ExampleParameters("only_python", tree_rc=1),
    ExampleParameters("parametric"),
    ExampleParameters("taskrunner"),
    ExampleParameters("trivial"),
]

examples_dir = Path("examples")


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_list(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "--list"],
        cwd=clean_bygg_tree / examples_dir / example.name,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.list_rc
    assert process.stdout == snapshot


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_list_directory(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "--list", "-C", examples_dir / example.name],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.list_rc
    assert process.stdout == snapshot


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_tree(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "--tree"],
        cwd=clean_bygg_tree / examples_dir / example.name,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.tree_rc
    assert process.stdout == snapshot


@pytest.mark.parametrize("example", examples, ids=lambda x: x.name)
def test_tree_directory(snapshot, clean_bygg_tree, example):
    process = subprocess.run(
        ["bygg", "--tree", "-C", examples_dir / example.name],
        cwd=clean_bygg_tree,
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == example.tree_rc
    assert process.stdout == snapshot


def test_dump_schema(snapshot):
    process = subprocess.run(
        ["bygg", "--dump-schema"],
        capture_output=True,
        encoding="utf-8",
    )
    assert process.returncode == 0
    assert process.stdout == snapshot



import argparse
from pathlib import Path
import shutil
import sys
import textwrap
from typing import Any, Generator, Set

from argcomplete.completers import DirectoriesCompleter
from argcomplete.finders import CompletionFinder

from bygg.output import output_plain


class ByggfileDirectoriesCompleter(DirectoriesCompleter):
    """
    A completer for directories that contain Bygg files.
    """

    def __call__(self, prefix, **kwargs) -> Generator[str, Any, None]:
        directories = super().__call__(prefix, **kwargs)
        byggfile_dirs = set()
        for dir in directories:
            byggfile_dirs |= {
                str(b.parent)
                for b in Path(dir).rglob("Byggfile.*")
                if b.suffix in {".py", ".yml"}
            }
        for f in sorted(list(byggfile_dirs)):
            yield f


def construct_already_there(
    parser: argparse.ArgumentParser | Any, options: Set[str]
) -> Set[str]:
    """
    Construct a set of options that are already present among the command line
    arguments. Adds both the short and long versions if they exist.
    """
    already_there_set: Set[str] = set()
    for action in parser._get_optional_actions():
        option_strings = set(action.option_strings)
        if option_strings & options:
            already_there_set.update(option_strings)
    return already_there_set


singular_options = {
    "--completions",
    "--dump-schema",
    "--help",
    "--version",
    "-h",
    "-v",
}

directory_option = {"-C", "--directory"}


class ByggCompletionFinder(CompletionFinder):
    """
    Override _get_completions because we want custom completion patterns that are not
    expressed in argparse.
    """

    def _get_completions(
        self, comp_words, cword_prefix, cword_prequote, last_wordbreak_pos
    ) -> list[str]:
        comp_words_set = set(comp_words)

        # Stop completions if we already have a "singular" option, i.e. an option that
        # should not be followed by anything else.
        if comp_words_set & singular_options:
            return []

        # Get completions from argparse.
        completions = set(
            super()._get_completions(
                comp_words, cword_prefix, cword_prequote, last_wordbreak_pos
            )
        )

        # If the last word is a directory, the default completions will be from the
        # directories completer, so return these.
        if comp_words[-1] in directory_option:
            return list(completions)

        action_completions = list(filter(lambda x: not x.startswith("-"), completions))
        already_there = construct_already_there(self._parser, comp_words_set)

        if comp_words_set & {"--tree", "--clean"}:
            # Actions if we have them.
            if action_completions and not cword_prefix.startswith("-"):
                return action_completions
            return list(completions & directory_option - already_there)

        if comp_words_set & {"--list"}:
            # Only complete directories.
            return list(completions & directory_option - already_there)

        # Return action completions if they exist and user has not entered a '-'.
        if action_completions and not cword_prefix.startswith("-"):
            return action_completions

        return list(completions - already_there)


def do_completion(parser: argparse.ArgumentParser):
    completer = ByggCompletionFinder()
    completer(parser)


def generate_shell_completions(print_and_exit: bool):
    bygg_bin = Path(sys.argv[0])
    bygg_completions = bygg_bin.resolve().parent / "_bygg_completions.sh"
    if not bygg_completions.exists():
        import argcomplete.shell_integration

        with open(bygg_completions, "w") as f:
            f.write(argcomplete.shell_integration.shellcode([str(bygg_bin.name)]))

    if print_and_exit:
        terminal_cols, _ = shutil.get_terminal_size()
        indent_string = "# "
        width = min(terminal_cols, 80) - len(indent_string)
        description = f"""\
            A shell completions script has been generated in {bygg_completions}. It will
            pick up the bygg that is in PATH. Add the following line to your shell's
            startup script to load completions:
        """
        output_plain(
            textwrap.indent(
                textwrap.fill(
                    textwrap.dedent(description),
                    width=width,
                    break_long_words=False,
                    break_on_hyphens=False,
                ),
                indent_string,
            )
        )
        output_plain(f"\nsource {bygg_completions}")

        sys.exit(0)



# PYTHON_ARGCOMPLETE_OK

import argparse
from dataclasses import dataclass
import os
import shutil
import stat
import subprocess
import sys
import textwrap
import time
from typing import Any, List

from bygg.apply_configuration import (
    apply_configuration,
    register_actions_from_configuration,
)
from bygg.argument_unparsing import unparse_args
from bygg.completions import (
    ByggfileDirectoriesCompleter,
    do_completion,
    generate_shell_completions,
)
from bygg.configuration import (
    PYTHON_INPUTFILE,
    YAML_INPUTFILE,
    ByggFile,
    dump_schema,
    read_config_file,
)
from bygg.job_output import output_job_logs
from bygg.output import (
    TerminalStyle as TS,
)
from bygg.output import (
    output_error,
    output_info,
    output_ok,
    output_plain,
    output_warning,
)
from bygg.runner import ProcessRunner
from bygg.scheduler import Scheduler
from bygg.status_display import (
    failed_checks,
    on_job_status,
    on_runner_status,
    output_check_results,
)
from bygg.system_helpers import change_dir
from bygg.tree import display_tree


@dataclass
class ByggContext:
    """Container for various state"""

    runner: ProcessRunner
    scheduler: Scheduler
    configuration: ByggFile


def get_job_count_limit():
    try:
        # Use os.sched_getaffinity where available (on U**X):
        # https://stackoverflow.com/a/55423170
        return len(os.sched_getaffinity(0))
    except AttributeError:
        count = os.cpu_count()
        assert count is not None
        return count


def init_bygg_context(configuration: ByggFile):
    scheduler = Scheduler()
    runner = ProcessRunner(scheduler)

    # Set up status listeners
    runner.job_status_listener = on_job_status
    runner.runner_status_listener = on_runner_status

    return ByggContext(runner, scheduler, configuration)


def build(
    ctx: ByggContext,
    actions: List[str],
    job_count: int | None,
    always_make: bool,
    check: bool,
) -> bool:
    """
    actions: The actions to build.

    job_count: The number of jobs to run simultaneously. None means to use the number of
    available cores.

    always_make: If True, all actions will be built, even if they are up to date.

    check: If True, apply various checks:

    * Check that the inputs and outputs of all actions will be checked for consistency.
      A job that runs later must not have files as output that are inputs to a job that
      runs earlier.
    """
    try:
        max_workers = get_job_count_limit() if job_count is None else job_count

        for action in actions:
            t1 = time.time()
            output_info(f"Building action '{action}':")

            ctx.scheduler.start_run(
                action,
                always_make=always_make,
                check=check,
            )
            status = ctx.runner.start(max_workers)
            ctx.scheduler.shutdown()

            if status:
                output_ok(f"Action '{action}' completed in {time.time() - t1:.2f} s.")
            else:
                output_error(
                    f"Action '{action}' failed after {time.time() - t1:.2f} s."
                )
                output_job_logs(ctx.runner.failed_jobs)
                return False

    except KeyboardInterrupt:
        output_warning("\nBuild was interrupted by user.")
        return False
    except KeyError as e:
        output_error(f"Error: Action '{e}' not found.")
        return False
    finally:
        ctx.scheduler.shutdown()

    if check and failed_checks:
        return output_check_results()

    return True


def clean(ctx: ByggContext, actions: List[str]):
    try:
        for action in actions:
            output_info(f"Cleaning action '{action}':")
            ctx.scheduler.prepare_run(action)
            for job_name in ctx.scheduler.job_graph.get_all_jobs():
                job = ctx.scheduler.build_actions.get(job_name, None)
                if job is None:
                    continue
                for output in job.outputs:
                    try:
                        s = os.stat(output)
                        if stat.S_ISREG(s.st_mode):
                            os.unlink(output)
                        elif stat.S_ISDIR(s.st_mode):
                            output_info(f"Removing directory: {output}")
                            shutil.rmtree(output)
                    except FileNotFoundError:
                        pass
            ctx.scheduler.shutdown()
    except KeyboardInterrupt:
        output_warning("Build was interrupted by user.")
        return False
    except KeyError as e:
        output_error(f"Error: Action '{e}' not found.")
        return False
    finally:
        ctx.scheduler.shutdown()

    return True


@dataclass
class EntryPoint:
    name: str
    description: str


NO_DESCRIPTION = "No description"


def get_entrypoints(ctx: ByggContext) -> List[EntryPoint]:
    return [
        EntryPoint(x.name, x.description or NO_DESCRIPTION)
        for x in ctx.scheduler.build_actions.values()
        if x.is_entrypoint
    ] or [
        EntryPoint(x.name, x.description or NO_DESCRIPTION)
        for x in ctx.configuration.actions
        if x.is_entrypoint
    ]


list_actions_style = "B"


def list_actions(ctx: ByggContext) -> bool:
    entrypoints = get_entrypoints(ctx)

    if not entrypoints:
        program_name = os.path.basename(sys.argv[0])
        output_error("Loaded build files but no entrypoints were found.")
        output_error(f"Type `{program_name} --help` for help.")
        return False

    terminal_cols, terminal_rows = shutil.get_terminal_size()
    output = [f"{TS.BOLD}Available actions:{TS.RESET}"]

    sorted_actions = sorted(entrypoints, key=lambda x: x.name)
    default_action_name = ctx.configuration.settings.default_action

    if default_action_name:
        default_action_list = [
            x for x in sorted_actions if x.name == default_action_name
        ]

        if default_action_list:
            default_action = default_action_list[0]
            default_action_name = default_action.name
            sorted_actions.remove(default_action)
            sorted_actions.insert(0, default_action)

    if list_actions_style == "A":
        output.append("")
        section_indent = 0
        separator = " : "
        max_name_width = max([len(x.name) for x in entrypoints])
        width = min(terminal_cols, 80)
        subsequent_indent = " " * (section_indent + max_name_width + len(separator))
        for action in sorted_actions:
            description = f"{TS.BOLD}{action.name: <{max_name_width}}{TS.RESET}{separator}{action.description}"
            output.extend(
                textwrap.wrap(
                    description,
                    width=width,
                    initial_indent=" " * section_indent,
                    subsequent_indent=subsequent_indent,
                )
            )
            output.append("")

    if list_actions_style == "B":
        output.append("")
        for action in sorted_actions:
            default_action_suffix = (
                " (default)" if action.name == default_action_name else ""
            )
            output.append(f"{TS.BOLD}{action.name}{default_action_suffix}{TS.RESET}")
            output.append(
                textwrap.fill(
                    action.description,
                    initial_indent="    ",
                    subsequent_indent="    ",
                )
            )
            output.append("")

    print("\n".join(output))

    return True


def print_version():
    import importlib.metadata

    if __package__:
        output_info(f"{__package__} {importlib.metadata.version(__package__)}")
    else:
        output_error("Could not determine package name.")


MAKE_COMPATIBLE_PANEL = "(Roughly) Make-compatible options"


def dispatcher():
    """
    A build tool written in Python, where all actions can be written in Python.
    """
    parser = create_argument_parser()
    do_completion(parser)

    args = parser.parse_args()

    generate_shell_completions(args.completions)

    if args.dump_schema:
        dump_schema()
        sys.exit(0)

    if args.version:
        print_version()
        sys.exit(0)

    directory = args.directory[0] if args.directory else None
    is_restarted_with_env = (
        args.is_restarted_with_env[0] if args.is_restarted_with_env else None
    )

    if directory and not is_restarted_with_env:
        directory_arg = args.directory[0]
        output_info(f"Entering directory '{directory_arg}'")
        os.chdir(directory_arg)

    if not os.path.isfile(PYTHON_INPUTFILE) and not os.path.isfile(YAML_INPUTFILE):
        output_error("No build files found.")
        sys.exit(1)

    configuration = read_config_file()
    ctx = init_bygg_context(configuration)

    try:
        action_partitions = (
            partition_actions(configuration, args.actions) if configuration else None
        )
    except KeyError as e:
        output_plain(
            f"Could not find action {TS.BOLD}{e}{TS.RESET}. List available actions with {TS.BOLD}--list{TS.RESET}."
        )
        sys.exit(1)

    if not action_partitions:
        if os.path.isfile(PYTHON_INPUTFILE):
            # No configuration file, so load the Python build file directly.
            apply_configuration(configuration, None, None)
            default_action = configuration.settings.default_action
            actions = (
                args.actions
                if args.actions
                else [default_action]
                if default_action
                else []
            )
            status = do_dispatch(ctx, args, actions)
        else:
            status = list_actions(ctx)

        if status:
            sys.exit(0)
        sys.exit(1)

    # Special case for --list here, since it shouldn't start loading the environments:
    if args.list:
        status = list_actions(ctx)
        sys.exit(0) if status else sys.exit(1)

    # Execute each action partition within the correct environment:

    for partition in action_partitions:
        env = partition.environment_name
        # Check if we should restart with another Python interpreter (e.g. from a
        # virtualenv):
        restart_with = apply_configuration(configuration, env, is_restarted_with_env)

        if restart_with is not None and not is_restarted_with_env:
            exec_list = [restart_with, *partition.actions] + unparse_args(
                parser, args, drop=["actions"]
            )
            if partition.environment_name:
                exec_list += ["--is_restarted_with_env", partition.environment_name]

            try:
                process = subprocess.run(exec_list, encoding="utf-8")
                if process.returncode != 0:
                    sys.exit(process.returncode)
            except FileNotFoundError:
                output_error(f"Error: Could not restart with '{restart_with}'.")
                output_warning(
                    "Please make sure that bygg is in your pip requirements list for this environment."
                )
                sys.exit(1)
        else:
            status = do_dispatch(ctx, args, partition.actions)
            if not status:
                sys.exit(1)


def do_dispatch(ctx: ByggContext, args: argparse.Namespace, actions: List[str]) -> bool:
    # Analysis implies building:
    always_make = args.always_make or args.check
    if args.clean:
        status = clean(ctx, actions)
    elif args.list:
        list_actions(ctx)
        status = True
    elif not actions:
        output_error("No actions specified and no default action is defined.\n")
        list_actions(ctx)
        status = False
    elif args.tree:
        status = display_tree(ctx.scheduler, actions)
    else:
        jobs = int(args.jobs) if args.jobs else None
        status = build(ctx, actions, jobs, always_make, args.check)

    return status


@dataclass
class ActionPartition:
    """
    environment_name: The name of the environment that the actions should be run in.
    None means the implicit default environment, i.e. typically the base process.

    actions: The actions that should be run in the environment.
    """

    environment_name: str | None
    actions: List[str]


def partition_actions(
    configuration: ByggFile,
    actions: List[str] | None,
) -> List[ActionPartition] | None:
    """
    Partition the actions into groups that should be run in the same environment. Only
    partition the given actions that also exist in the configuration file. This is to
    not have to load the Python build files for all the environments, since installing
    their respective requiements can take a while.

    Parameters
    ----------
    configuration : ByggFile
        The configuration file.
    actions : List[str] | None
        The actions that should be run. If None, a partition will be created, resolved
        to the default action.

    Returns
    -------
    List[ActionPartition] | None
        A list of ActionPartition objects, each representing a group of actions that
        should be run in the same environment. Returns None if there are no actions in
        the configuration.
    """
    if not configuration.actions:
        return None

    resolved_actions = actions if actions else []

    if not resolved_actions and configuration.settings.default_action is not None:
        # Resolve to default action:
        resolved_actions += [configuration.settings.default_action]
        return [ActionPartition("default", resolved_actions)]

    # Put consecutive actions with the same environment in the same partition:
    action_partitions = []
    action_dict = {a.name: a for a in configuration.actions}
    current_partition: ActionPartition | None = None
    for action in [action_dict[a] for a in resolved_actions]:
        if (
            current_partition is None
            or current_partition.environment_name != action.environment
        ):
            current_partition = ActionPartition(action.environment, [action.name])
            action_partitions.append(current_partition)
        else:
            current_partition.actions.append(action.name)

    return action_partitions


def entrypoint_completions(prefix, parsed_args: argparse.Namespace, **kwargs):
    import textwrap

    # Handle -C/--directory:
    new_dir = parsed_args.directory[0] if parsed_args.directory else None

    with change_dir(new_dir):
        configuration = read_config_file()
        ctx = init_bygg_context(configuration)

        if configuration.environments:
            # We don't want to install environments while completing, so only load the
            # actions from the static configuration file:
            register_actions_from_configuration(ctx.configuration, None)
        else:
            apply_configuration(ctx.configuration, None, None)

        entrypoints = get_entrypoints(ctx)
        # Exlude already added entrypoints:
        eligible_entrypoints = [
            x for x in entrypoints if x.name not in parsed_args.actions
        ]
        # Fill the description really wide to get it combined to a single line with
        # spaces dealt with correctly:
        return {
            x.name: textwrap.fill(x.description, 7000) for x in eligible_entrypoints
        }


def create_argument_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""
A build tool written in Python, where all actions can be written in Python.

Build the default action:
 %(prog)s

Build ACTION:
 %(prog)s ACTION

Clean ACTION:
 %(prog)s ACTION --clean

List available actions:
 %(prog)s --list
""",
    )

    # Use Any to get around type checking for argcomplete:
    arg: Any = parser.add_argument(
        "actions",
        nargs="*",
        default=None,
        help="Entrypoint actions to operate on.",
    )
    arg.completer = entrypoint_completions

    parser.add_argument(
        "-v", "--version", action="store_true", help="Show version string and exit."
    )

    parser.add_argument(
        "--is_restarted_with_env",
        nargs=1,
        type=str,
        default=None,
        help=argparse.SUPPRESS,
    )
    # Commands that operate on the build setup:
    build_setup_wrapper_group = parser.add_argument_group(
        "Commands that operate on the build setup"
    )  # add_mutually_exclusive_group doesn't accept a title, so wrap it in a regular group.
    build_setup_group = build_setup_wrapper_group.add_mutually_exclusive_group()
    build_setup_group.add_argument(
        "--clean",
        action="store_true",
        help="Clean the outputs of the specified actions.",
    )
    build_setup_group.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="List available actions.",
    )
    build_setup_group.add_argument(
        "--tree",
        action="store_true",
        help="Display the dependency tree starting from the specified action(s).",
    )
    # Some arguments inspired by Make:
    make_group = parser.add_argument_group("Make-like arguments")
    arg = make_group.add_argument(
        "-C",
        "--directory",
        nargs=1,
        type=str,
        default=None,
        help="Change to the specified directory.",
    )
    arg.completer = ByggfileDirectoriesCompleter()

    make_group.add_argument(
        "-j",
        "--jobs",
        nargs="?",
        type=int,
        default=None,
        help="Specify the number of jobs to run simultaneously. None means to use the number of available cores.",
    )
    make_group.add_argument(
        "-B",
        "--always-make",
        action="store_true",
        help="Always build all actions.",
    )

    # Analyse and verify:
    analyse_group = parser.add_argument_group(
        "Analyse and verify",
        "Arguments in this group will add more analysis to the build process. Actions will be built and the analysis result will be reported.",
    )
    analyse_group.add_argument(
        "--check",
        action="store_true",
        help="Perform various checks on the action tree. Implies -B",
    )

    # Meta arguments:
    meta_group = parser.add_argument_group("Meta arguments")
    meta_group.add_argument(
        "--dump-schema",
        action="store_true",
        help="Generate a JSON Schema for the Byggfile.yml files. The schema will be printed to stdout.",
    )
    meta_group.add_argument(
        "--completions",
        action="store_true",
        help="Output instructions for how to set up shell completions via the shell's startup script.",
    )

    return parser


def main():
    try:
        return dispatcher()
    except KeyboardInterrupt:
        output_warning("Interrupted by user. Aborting.")
        return 1


if __name__ == "__main__":
    main()



import os
from pathlib import Path
import shutil
import subprocess
import sys
import textwrap
from typing import List

from bygg.action import Action
from bygg.configuration import PYTHON_INPUTFILE, ByggFile, Environment
from bygg.digest import calculate_string_digest
from bygg.output import output_error, output_info, output_plain
from bygg.util import create_shell_command


def calculate_environment_hash(environment: Environment) -> str:
    requirements: List[str] = []
    requirements.append(environment.shell)
    if environment.inputs:
        for input in environment.inputs:
            with open(input, "r") as f:
                requirements += f.readlines()
    return calculate_string_digest(" ".join(requirements))


def setup_environment(environment: Environment):
    venv_path = Path(
        environment.venv_directory if environment.venv_directory else ".venv"
    )

    environment_hash = calculate_environment_hash(environment)
    environment_hash_file = venv_path / "bygg_environment_hash.txt"
    if environment_hash_file.exists():
        with open(environment_hash_file, "r") as f:
            if f.read() == environment_hash:
                return True
        environment_hash_file.unlink()

    # TODO should we remove the whole venv or just trust pip to do the right thing? For
    # now, remove the venv.

    if venv_path.exists():
        output_info(f"Replacing venv at {venv_path}")
        shutil.rmtree(venv_path)

    output_info("Setting up environment")
    process = subprocess.run(
        environment.shell,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
    )

    if process.returncode != 0:
        output_error("Error while creating virtual environment:")
        output_plain(process.stdout)
        sys.exit(1)

    with open(environment_hash_file, "w") as f:
        f.write(environment_hash)

    return True


def should_restart_with(environment: Environment) -> str | None:
    venv_bin_path = Path(environment.venv_directory) / "bin"

    # # TODO: we could check here if the new venv is using the same Python binary that
    # # we're already running under, and if so just activate the venv.

    is_relative = Path(sys.executable).is_relative_to(venv_bin_path.resolve())
    if not is_relative:
        return str((venv_bin_path / "bygg"))
    return None


def register_actions_from_configuration(
    configuration: ByggFile, is_restarted_with_env: str | None
):
    for action in configuration.actions:
        if is_restarted_with_env and action.environment != is_restarted_with_env:
            continue

        shell_command = (
            create_shell_command(action.shell, action.message) if action.shell else None
        )
        Action(
            action.name,
            description=action.description,
            is_entrypoint=bool(action.is_entrypoint),
            inputs=action.inputs,
            outputs=action.outputs,
            dependencies=action.dependencies,
            command=shell_command,
        )


def apply_configuration(
    configuration: ByggFile,
    environment_name: str | None,
    is_restarted_with_env: str | None,
) -> str | None:
    """Returns the path of the bygg install to restart with if a restart is needed."""

    environment = (
        configuration.environments[environment_name]
        if environment_name and environment_name in configuration.environments
        else None
    )

    # Check if we need to restart to run in a different environment:

    if not is_restarted_with_env and environment:
        setup_environment(environment)

        restart_with = should_restart_with(environment)
        if restart_with is not None:
            return restart_with

    # Now set up the actions for the current environment:
    register_actions_from_configuration(configuration, is_restarted_with_env)

    # Evaluate the Python build file:

    python_build_file = environment.byggfile if environment else PYTHON_INPUTFILE
    load_python_build_file(python_build_file)
    return None


def load_python_build_file(build_file: str):
    # modify load path to make the current directory importable
    preamble = """\
        import os
        import sys
        sys.path.insert(0, str(os.path.realpath('.')))

        """

    if os.path.isfile(build_file):
        with open(build_file, "r") as f:
            exec(textwrap.dedent(preamble) + f.read(), globals())
